/*-----------------------------------------------------------------------*/
/* Program: Stream                                                       */
/* Revision: $Id: caspur_stream.c,v 1.0 2005/7/18 luca.ferraro@caspur.it */
/* Revision: $Id: stream.c,v 5.4 2005/02/08 03:05:20 mccalpin Exp mccalpin $ */
/* Original code developed by John D. McCalpin                           */
/* Programmers: John D. McCalpin                                         */
/*              Joe R. Zagar                                             */
/*                                                                       */
/* This program measures memory transfer rates in MB/s for simple        */
/* computational kernels coded in C.                                     */
/*-----------------------------------------------------------------------*/
/* Copyright 1991-2005: John D. McCalpin                                 */
/*-----------------------------------------------------------------------*/
/* License:                                                              */
/*  1. You are free to use this program and/or to redistribute           */
/*     this program.                                                     */
/*  2. You are free to modify this program for your own use,             */
/*     including commercial use, subject to the publication              */
/*     restrictions in item 3.                                           */
/*  3. You are free to publish results obtained from running this        */
/*     program, or from works that you derive from this program,         */
/*     with the following limitations:                                   */
/*     3a. In order to be referred to as "STREAM benchmark results",     */
/*         published results must be in conformance to the STREAM        */
/*         Run Rules, (briefly reviewed below) published at              */
/*         http://www.cs.virginia.edu/stream/ref.html                    */
/*         and incorporated herein by reference.                         */
/*         As the copyright holder, John McCalpin retains the            */
/*         right to determine conformity with the Run Rules.             */
/*     3b. Results based on modified source code or on runs not in       */
/*         accordance with the STREAM Run Rules must be clearly          */
/*         labelled whenever they are published.  Examples of            */
/*         proper labelling include:                                     */
/*         "tuned STREAM benchmark results"                              */
/*         "based on a variant of the STREAM benchmark code"             */
/*         Other comparable, clear and reasonable labelling is           */
/*         acceptable.                                                   */
/*     3c. Submission of results to the STREAM benchmark web site        */
/*         is encouraged, but not required.                              */
/*  4. Use of this program or creation of derived works based on this    */
/*     program constitutes acceptance of these licensing restrictions.   */
/*  5. Absolutely no warranty is expressed or implied.                   */
/*-----------------------------------------------------------------------*/
# include <stdlib.h>
# include <stdio.h>
# include <math.h>
# include <float.h>
# include <limits.h>
# include <sys/time.h>

/* INSTRUCTIONS:
 *
 *	1) Stream requires a good bit of memory to run.  Adjust the
 *          value of 'N' (below) to give a 'timing calibration' of 
 *          at least 20 clock-ticks.  This will provide rate estimates
 *          that should be good to about 5% precision.
 */

#ifndef N
# define N	2000000
#endif

#ifndef NTIMES 
# define NTIMES	20
#endif


#ifndef OFFSET
# define OFFSET	0
#endif

static  int     size, iterations, offset;


/*
 *	3) Compile the code with full optimization.  Many compilers
 *	   generate unreasonably bad code before the optimizer tightens
 *	   things up.  If the results are unreasonably good, on the
 *	   other hand, the optimizer might be too smart for me!
 *
 *         Try compiling with:
 *               cc -O stream_omp.c -o stream_omp
 *
 *         This is known to work on Cray, SGI, IBM, and Sun machines.
 *
 *
 *	4) Mail the results to mccalpin@cs.virginia.edu
 *	   Be sure to include:
 *		a) computer hardware model number and software revision
 *		b) the compiler flags
 *		c) all of the output from the test case.
 * Thanks!
 *
 */

# define HLINE "-------------------------------------------------------------\n"

# ifndef MIN
# define MIN(x,y) ((x)<(y)?(x):(y))
# endif
# ifndef MAX
# define MAX(x,y) ((x)>(y)?(x):(y))
# endif

/* data array for streaming kernel tests */
static double *a, *b, *c;

static double	avgtime[4] = {0}, maxtime[4] = {0},
		mintime[4] = {FLT_MAX,FLT_MAX,FLT_MAX,FLT_MAX};

static char	*label[4] = {"Copy:      ", "Scale:     ",
    "Add:       ", "Triad:     "};


double mysecond();


int main( int argc, char **argv)
{
    int			quantum, checktick();
    int			BytesPerWord;
    register int	j, k;
    double		scalar, t, *times[4];
    double      *avrtime2[4], rms[4];
    double	    bytes[4];

    size = N;
    iterations = NTIMES;
    offset = OFFSET;

    printf(HLINE);
    printf("\nMemory required for current stream test %8.3f Mbyte\n",
              ((double) 3 * size * sizeof(double))/1024/1024 );    

    /* Allocate and size stream data arrays */
    a = (double *) malloc ( (size + offset) * sizeof(double) );
    b = (double *) malloc ( (size + offset) * sizeof(double) );
    c = (double *) malloc ( (size + offset) * sizeof(double) );

    if ( c == NULL || b == NULL || a == NULL)
    {
        fprintf(stderr,"\nCannot allocate so much memory !\n");    
        exit(EXIT_FAILURE);
    }
    
    /* Allocate array to store bench results for each iteration *
     * and do some statistic on it                               */
    for (k = 0; k < 4; k++)
    {
        times[k] = (double *) malloc ( iterations * sizeof(double) );
        avrtime2[k] = (double *) malloc ( iterations * sizeof(double) );

        /* zeroing initial values */
        for (j = 0; j < iterations; j++)
            avrtime2[k][j] = 0.0;    
        rms[k] = 0.0;
    }

    bytes[0] =  2 * sizeof(double) * size;
    bytes[1] =  2 * sizeof(double) * size;
    bytes[2] =  3 * sizeof(double) * size;
    bytes[3] =  3 * sizeof(double) * size;




    /* --- SETUP --- determine precision and check timing --- */
    printf(HLINE);
    BytesPerWord = sizeof(double);
    printf("This system uses %d bytes per DOUBLE PRECISION word.\n",
	BytesPerWord);

    printf(HLINE);
    printf("Each test will be run %d times.\n", iterations);

#ifdef _OPENMP
    printf(HLINE);
#pragma omp parallel private(k)
    {
        k = omp_get_num_threads();
        printf ("Number of Threads requested = %i\n",k);
    }
#endif

    /* Get initial value for system clock. */
#pragma omp parallel for
    for (j = 0; j < size; j++) 
    {
	    a[j] = 1.0;
	    b[j] = 2.0;
	    c[j] = 0.0;
	}

    printf(HLINE);

    if  ( (quantum = checktick()) >= 1) 
	printf("Your clock granularity/precision appears to be "
	    "%d microseconds.\n", quantum);
    else
	printf("Your clock granularity appears to be "
	    "less than one microsecond.\n");

    t = mysecond();
#pragma omp parallel for
    for (j = 0; j < size; j++)
	    a[j] = 2.0E0 * a[j];
    t = 1.0E6 * (mysecond() - t);

    printf("Each test below will take on the order"
	" of %d microseconds.\n", (int) t  );
    printf("   (= %d clock ticks)\n", (int) (t/quantum) );
    printf("Increase the size of the arrays if this shows that\n");
    printf("you are not getting at least 20 clock ticks per test.\n");

    printf(HLINE);

    printf("WARNING -- The above is only a rough guideline.\n");
    printf("For best results, please be sure you know the\n");
    printf("precision of your system timer.\n");
    printf(HLINE);
    




    /*	--- MAIN LOOP --- repeat test cases iterations times --- */

    scalar = 3.0;
    for (k = 0; k < iterations; k++)
	{
	    times[0][k] = mysecond();
#ifdef TUNED
        tuned_STREAM_Copy();
#else
#pragma omp parallel for
	    for (j = 0; j < size; j++)
	        c[j] = a[j];
#endif
	    times[0][k] = mysecond() - times[0][k];
        avrtime2[0][k] = times[0][k] * times[0][k];
	
	    times[1][k] = mysecond();
#ifdef TUNED
        tuned_STREAM_Scale(scalar);
#else
#pragma omp parallel for
	    for (j = 0; j < size; j++)
	        b[j] = scalar*c[j];
#endif
	    times[1][k] = mysecond() - times[1][k];
        avrtime2[1][k] = times[1][k] * times[1][k];
	
	    times[2][k] = mysecond();
#ifdef TUNED
        tuned_STREAM_Add();
#else
#pragma omp parallel for
	    for (j = 0; j < size; j++)
	        c[j] = a[j]+b[j];
#endif
	    times[2][k] = mysecond() - times[2][k];
        avrtime2[2][k] = times[2][k] * times[2][k];
	
	    times[3][k] = mysecond();
#ifdef TUNED
        tuned_STREAM_Triad(scalar);
#else
#pragma omp parallel for
	    for (j = 0; j < size; j++)
	        a[j] = b[j]+scalar*c[j];
#endif
    	times[3][k] = mysecond() - times[3][k];
        avrtime2[3][k] = times[3][k] * times[3][k];
	}

    /*	--- SUMMARY --- */

    for (k = 1; k < iterations; k++) /* note -- skip first iteration */
	{
	    for (j = 0; j < 4; j++)
	    {
	        avgtime[j] = avgtime[j] + times[j][k];
	        mintime[j] = MIN(mintime[j], times[j][k]);
	        maxtime[j] = MAX(maxtime[j], times[j][k]);
            /* used to compute root mean square deviations */
	        rms[j] = rms[j] + avrtime2[j][k] * avrtime2[j][k]; 
	    }
	}
    
    printf("Function    Rate (MB/s)   Rms (MB/s)   Avg time    Rms time    Min time    Max time\n");
    for (j = 0; j < 4; j++) 
    {
	    avgtime[j] = avgtime[j]/(double)(iterations - 1);
        rms[j]    =  sqrt( rms[j] / (double)(iterations - 1) );

	    printf("%s%11.4f %11.4f %11.4f %11.4f %11.4f %11.4f\n", label[j],
	       1.0E-06 * bytes[j]/avgtime[j],
	       1.0E-06 * bytes[j]/(avgtime[j]*avgtime[j]) * rms[j],
	       avgtime[j],
           rms[j],
	       mintime[j],
	       maxtime[j]);
    }
    printf(HLINE);

    return 0;
}











# define	M	20

int
checktick()
{
    int		i, minDelta, Delta;
    double	t1, t2, timesfound[M];

/*  Collect a sequence of M unique time values from the system. */

    for (i = 0; i < M; i++) {
	t1 = mysecond();
	while( ((t2=mysecond()) - t1) < 1.0E-6 )
	    ;
	timesfound[i] = t1 = t2;
	}

/*
 * Determine the minimum difference between these M values.
 * This result will be our estimate (in microseconds) for the
 * clock granularity.
 */

    minDelta = 1000000;
    for (i = 1; i < M; i++) {
	Delta = (int)( 1.0E6 * (timesfound[i]-timesfound[i-1]));
	minDelta = MIN(minDelta, MAX(Delta,0));
	}

   return(minDelta);
}



/* A gettimeofday routine to give access to the wall
   clock timer on most UNIX-like systems.  */

#include <sys/time.h>

double mysecond()
{
        struct timeval tp;
        struct timezone tzp;
        int i;

        i = gettimeofday(&tp,&tzp);
        return ( (double) tp.tv_sec + (double) tp.tv_usec * 1.e-6 );
}

